import tempfile
import subprocess as sp
import sys
from tqdm import tqdm
import os
import logging


# Central class of microRNA objects
class Rna(object):
    def __init__(self, name, chromosome, start, end, strand, seq, bit):
        # miRNA identifier
        self.name = name
        # chromosome that the miRNA is located on
        self.chromosome = chromosome
        # start position of the pre-miRNA
        self.start = int(start)
        # end position of the pre-miRNA
        self.end = int(end)
        # sense (+) or anti-sense (-) strand
        self.strand = strand
        # nucleotide sequence of the pre-miRNA
        self.seq = seq.replace('U', 'T')
        # reference bit score that miRNA receives by its own
        # covariance model
        self.bit = bit


def find_cm_refbit(mirid, seq, model, cpu):
    """
    Obtain the reference bit score for each miRNA by applying it to its own covariance model.
    """
    logger = logging.getLogger('ncortho')
    logger.setLevel(level=logging.DEBUG)
    with tempfile.NamedTemporaryFile(mode='w+') as fp:
        fp.write('>{0}\n{1}'.format(mirid, seq))
        fp.seek(0)
        cms_command = f'cmsearch --cpu {cpu} -E 0.01 --noali {model} {fp.name}'
        res = sp.run(cms_command, shell=True, capture_output=True)
    if res.returncode != 0:
        raise sp.SubprocessError(res.stderr.decode("utf-8"))

    for line in res.stdout.decode('utf-8').split('\n'):
        # print(line)
        if line.startswith('  '):
            highest_score = float(line.split()[3])
            return highest_score
    # if no results return topscore 0
    logger.info('Self bit score not applicable, setting threshold to 0')
    return 0.0


def find_phmm_refbit(rnaid, seq, model, cpu):
    logger = logging.getLogger('ncortho')
    logger.setLevel(level=logging.DEBUG)

    call = sp.Popen(
        f'nhmmer -E 0.01 --cpu {cpu} --noali {model} "-"',
        shell=True, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE, encoding='utf8'
        )
    res, err = call.communicate(f'>{rnaid}\n{seq}\n')
    if err:
        raise sp.SubprocessError(err)
    reslist = res.split('\n')
    for i, line in enumerate(reslist):
        if 'Evalue' in line:
            resline = reslist[i+2]
            topscore = float(resline.split()[1])
            return topscore
    # if no results return topscore 0
    logger.info('Self bit score not applicable, setting threshold to 0')
    return 0.0


def rna_maker(mirpath, modeldir, phmm, cpu):
    logger = logging.getLogger('ncortho')
    logger.setLevel(level=logging.DEBUG)
    """
    Parses the miRNA data input file and returns a dictionary of Mirna objects.

    mirpath : Path to file with microRNA data.
    modeldir : Path to directory containing models.

    Returns : {'mirid': Mirna, ...}

    """
    with open(mirpath) as mirna_file:
        rna_data = [
            line.strip().split()[:6] for line in mirna_file if not line.startswith('#')
        ]

    mmdict = {}  # will be the return object
    print('# Calculating reference bit scores')
    for rna in tqdm(rna_data, file=sys.stdout):
        rnaid = rna[0]
        seq = rna[5]

        if phmm:
            model = '{0}/{1}.phmm'.format(modeldir, rnaid)
        else:
            model = '{0}/{1}.cm'.format(modeldir, rnaid)
        if not os.path.isfile(model):
            logger.warning(f'No model found for {rnaid}. Accepted are ".cm" and ".phmm"')
            mmdict[rnaid] = ''
            continue

        if phmm:
            top_score = find_phmm_refbit(rnaid, seq, model, cpu)
        else:
            top_score = find_cm_refbit(rnaid, seq, model, cpu)

        rna.append(top_score)

        # Create output
        mmdict[rnaid] = Rna(*rna)

    return mmdict


# Construct and calibrate a covariance model for a given ncRNA core set
# alignment in Stockholm format.
# The calibration step is computationally very expensive and should be
# performed on multiple CPU cores.
# TODO: Include license, author etc.

import argparse
import multiprocessing as mp
import os
import subprocess as sp
import sys


class CmConstructor(object):
    
    def __init__(self, alignment, outpath, name, cpu):
        self.alignment = alignment
        self.outpath = outpath
        self.name = name
        self.cpu = cpu
        self.model = '{0}/{1}.cm'.format(outpath, name)
    
    def construct(self):
        # print('# Constructing covariance model for {}.'.format(self.name), '')
        construct_command = (
            'cmbuild -F -n {0} {1}/{0}.cm {2}'
            .format(self.name, self.outpath, self.alignment)
        )
        sp.run(construct_command, shell=True, capture_output=True)
        # print('# Finished covariance model construction.')
    
    def calibrate(self):
        # print('# Calibrating covariance model for {}.'.format(self.name))
        calibrate_command = (
            'cmcalibrate --cpu {0} {1}'.format(self.cpu, self.model)
        )
        sp.run(calibrate_command, shell=True)
        # print('# Finished covariance model calibration.')


def create_cm(alignment, outpath, name, cpu):
    # Initiate covariance model construction and calibration.
    cmc = CmConstructor(alignment, outpath, name, cpu)
    # Construct the model.
    cmc.construct()
    # Calibrate the model.
    cmc.calibrate()


def create_phmm(alignment, outpath, name, cpu):
    model = f'{outpath}/{name}.phmm'
    cmd = f'hmmbuild -n {name} --informat stockholm --dna {model} {alignment}'
    res = sp.run(cmd, shell=True, capture_output=True)
    res.check_returncode()


# def main():
#
#     # Parse command-line arguments.
#     parser = argparse.ArgumentParser(
#         prog='python createcm.py', description='covariance model construction'
#     )
#     # "cpu", use maximum number of available CPUs unless specified otherwise.
#     parser.add_argument(
#         '-c', '--cpu', metavar='int', type=int,
#         help='number of CPU cores to use', nargs='?',
#         const=mp.cpu_count(), default=mp.cpu_count()
#     )
#     # Path to the desired output folder.
#     parser.add_argument(
#         '-o', '--output', metavar='<path>', type=str,
#         help='path to the output folder'
#     )
#     # Path to the input alignment
#     parser.add_argument(
#         '-a', '--alignment', metavar='<.sto>', type=str,
#         help='path to input alignment'
#     )
#
#     # Show help when no arguments are added.
#     if len(sys.argv) == 1:
#         parser.print_help()
#         sys.exit(1)
#     else:
#         args = parser.parse_args()
#
#     # Check if computer provides the desired number of cores.
#     available_cpu = mp.cpu_count()
#     if args.cpu > available_cpu:
#         raise ValueError('The provided number of CPU cores is higher than the number available on this system')
#     else:
#         cpu = args.cpu
#
#     # Check if alignment file exists.
#     if not os.path.isfile(args.alignment):
#         raise ValueError('Invalid path to alignment file')
#     else:
#         alignment = args.alignment
#
#     output = args.output
#     name = alignment.split('/')[-1].split('.')[0]
#
#     # Check if the output folder exists.
#     if not os.path.isdir(output):
#         mkdir_cmd = 'mkdir -p {}'.format(output)
#         sp.call(mkdir_cmd, shell=True)
#
#     # Initiate covariance model construction and calibration.
#     cmc = CmConstructor(alignment, output, name, cpu)
#     # Construct the model.
#     cmc.construct()
#     # Calibrate the model.
#     cmc.calibrate()
#
#
# if __name__ == '__main__':
#     main()



import argparse
import sys
import os
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import yaml


def parse_yaml(path):
    p_dict = {}  # {<name>: {'orthologs': <path>, 'genome': <path>, 'annotation': <path>}}
    paths = []
    refdict = {}
    with open(path, 'r') as param_handle:
        params = yaml.load_all(param_handle, Loader=yaml.FullLoader)
        for entry in params:
            in_type = entry.pop('type')
            if in_type == 'reference':
                refdict['annotation'] = entry['annotation']
                refdict['geneset'] = entry['geneset']
            elif in_type == 'core':
                species = entry.pop('name')
                p_dict[species] = entry
            paths.extend([entry['annotation'], entry['geneset']])

        return p_dict, refdict, paths


def read_ortho_tsv(path):
    orthod = {}
    allhumanids = set()
    with open(path) as fh:
        for line in fh:
            if line.startswith('#'):
                continue
            human_protid, core_protid = line.strip().split()
            if not human_protid in orthod:
                orthod[human_protid] = []
            if not core_protid in orthod:
                orthod[core_protid] = []
            orthod[human_protid].append(core_protid)
            orthod[core_protid].append(human_protid)
            allhumanids.add(human_protid)
    return orthod, allhumanids


def geneorder_from_gff(path, proteinset):
    """
    Read gene order from GFF file. Will only collect order of genes for which an ortholog was searched

    returns:
        orderpercontig = {contig: [gene1, gene2, gene3], ...}
        contigperid = {gene1: contig, ...}
    """
    orderpercontig = {}
    contigperid = {}
    lastcontig = ''
    with open(path) as fh:
        for line in fh:
            if line.startswith('#'):
                continue
            data = line.strip().split('\t')
            if data[2] != 'CDS':
                continue
            contig = data[0]
            if contig != lastcontig:
                orderpercontig[contig] = []
                lastcontig = contig
            protid = data[-1].split('-')[1].split('.')[0]

            if orderpercontig[contig] and orderpercontig[contig][-1] == protid:  # do not add duplicates
                continue
            if protid in proteinset:
                orderpercontig[contig].append(protid)
                contigperid[protid] = contig
    return orderpercontig, contigperid


def geneset_from_file(path):
    if path.split('.')[-1] in ['fasta', 'fa', 'faa']:
        o = set()
        with open(path) as fh:
            for line in fh:
                if line.startswith('>'):
                    o.add(line.split('.')[0].replace('>', ''))
    else:
        with open(path) as fh:
            o = [line.strip() for line in fh if line]
            o = set(o)
    return o


def neighbors(protid, orderpercontig, contigperid, k):
    contig = contigperid[protid]
    orderlist = orderpercontig[contig]
    position = orderlist.index(protid)
    # print(contig, position)
    leftflank = orderlist[position - k:position]
    rightflank = orderlist[position + 1:position + k + 1]
    # print(leftflank)
    # print(rightflank)

    return leftflank, rightflank


def orthomapper(referencelist, orthodict):
    o = []
    for prot in referencelist:
        if prot not in orthodict:
            continue
        o.extend(orthodict[prot])
    return o


def jaccard_similarity(list1, list2):
    intersection = len(list(set(list1).intersection(list2)))
    union = (len(set(list1)) + len(set(list2))) - intersection
    if union == 0:
        return union
    else:
        return float(intersection) / union


def main():
    # Print header
    print('\n' + '#' * 38, flush=True)
    print('###' + ' ' * 32 + '###', flush=True)
    print('###   ncOrtho - core set synteny   ###', flush=True)
    print('###' + ' ' * 32 + '###', flush=True)
    print('#' * 38 + '\n', flush=True)

    # Parse command-line arguments
    # Define global variables
    parser = argparse.ArgumentParser(
        description=(
            'Build Covariance models of reference miRNAs from core set of orthologs.'
        )
    )
    parser._action_groups.pop()
    required = parser.add_argument_group('Required Arguments')
    optional = parser.add_argument_group('Optional Arguments')
    # input file path
    required.add_argument('-p', '--parameters', metavar='<path>', type=str, required=True,
                          help='Path to the parameters file in yaml format')
    # output folder
    required.add_argument(
        '-o', '--output', metavar='<path>', type=str, required=True,
        help='Path for the output folder'
    )
    optional.add_argument(
        '-k', '--max_anchor_dist', metavar='int', type=int,
        help='Number of additional genes to the left and right '
             'of the reference miRNA that are to be considered as syntenic anchors. (Default: 3)',
        nargs='?', const=3, default=3
    )
    # optional.add_argument(
    #     '--idtype', metavar='str', type=str,
    #     help='Choose the ID in the reference gff file that is '
    #          'compared to the IDs in the pairwise orthologs file (default:GeneID) '
    #          'Options: ID, Name, GeneID, CDS',
    #     nargs='?', const='ID=', default='ID'
    # )
    optional.add_argument(
        '--outformat', metavar='str', type=str,
        help='Choose format for output figures, as accepted by matplotlib package (Default: png)',
        nargs='?', const='png', default='png'
    )
    # Show help when no arguments are added.
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    else:
        args = parser.parse_args()


    core_paths, ref_paths, all_paths = parse_yaml(args.parameters)
    ref_protset = geneset_from_file(ref_paths['geneset'])
    ref_orderpercontig, ref_contigperid = geneorder_from_gff(ref_paths['annotation'], ref_protset)
    print('# Reading reference annotation done', flush=True)

    syn_col = []
    ortho_col = []
    for corespecies, pathdict in core_paths.items():
        print(corespecies)
        pairwise_orthos, allrefids = read_ortho_tsv(pathdict['orthologs'])
        ortho_col.append([corespecies, len(allrefids)])

        core_protset = geneset_from_file(pathdict['geneset'])
        core_orderpercontig, core_contigperid = geneorder_from_gff(pathdict['annotation'], core_protset)

        for refprot in allrefids:
            # check if an ortholog was found for the protein in the reference species
            if not refprot in pairwise_orthos:
                syn_col.append([corespecies, refprot, None])
                continue

            # find position and neighbors of reference protein
            if refprot not in ref_contigperid:  # rare cases of protein.rep.fa and genomic.gff not matching up
                continue
            ref_left, ref_right = neighbors(refprot, ref_orderpercontig, ref_contigperid, args.max_anchor_dist)

            # map neighboring genes to orthologs in core species
            ref_ortho_left = orthomapper(ref_left, pairwise_orthos)
            ref_ortho_right = orthomapper(ref_right, pairwise_orthos)

            # find position and neighbors of the (co-)ortholog to the reference protein in the core species
            coreproteins = pairwise_orthos[refprot]
            for coreprot in coreproteins:
                core_left, core_right = neighbors(coreprot, core_orderpercontig, core_contigperid, args.max_anchor_dist)

                reflist = ref_ortho_left + ref_ortho_right
                corelist = core_left + core_right

                ji = jaccard_similarity(reflist, corelist)
                syn_col.append([corespecies, refprot, ji])


    sns.set(rc={'figure.figsize': (6, 6), 'ytick.left': True, 'xtick.bottom': True}, font_scale=1.2, style='whitegrid')
    orth = pd.DataFrame(ortho_col, columns=['species', 'Number orthologs'])
    sns.barplot(data=orth, x='species', y='Number orthologs', edgecolor="0.3", linewidth=1.5)
    plt.xticks(rotation=45, ha='right')
    plt.xlabel('')
    plt.tight_layout()
    plt.savefig(f'{args.output}/ortholog_distribution.{args.outformat}')


    plt.figure(figsize=(6, 6))
    sns.set(rc={'figure.figsize': (6, 6), 'ytick.left': True, 'xtick.bottom': True}, font_scale=1.2, style='whitegrid')
    df = pd.DataFrame(syn_col, columns=['species', 'reference_protein', 'jaccard_index'])
    df.to_csv(f'{args.output}/conserved_synteny.tsv', sep='\t', index=False)
    sns.boxplot(data=df, x='species', y='jaccard_index', linewidth=1.5)
    plt.xlabel('')
    # plt.title(f'Flanking gene window size = {max_anchor_dist}')
    plt.ylabel(f'Jaccard Similarity (k = {args.max_anchor_dist})')
    plt.xticks(rotation=45, ha='right')
    plt.tight_layout()
    plt.savefig(f'{args.output}/anchor_conservation.{args.outformat}')

    print(f'# Finished. Output created at: {args.output}')

if __name__ == '__main__':
    main()



import glob
import subprocess as sp
import argparse


def check_blastdb(db_path):
    file_extensions = ['nhr', 'nin', 'nsq']
    for fe in file_extensions:
        files = glob.glob(f'{db_path}*{fe}')
        if not files:
            # At least one of the BLAST db files is not existent
            return False
    return True


def make_blastndb(inpath, outpath):
    db_command = 'makeblastdb -in {} -out {} -dbtype nucl'.format(inpath, outpath)
    sp.run(db_command, shell=True, capture_output=True)


def vprint(s, verbose):
    if verbose:
        print(s, flush=True)


def write_output(outlist, path):
    with open(path, 'w') as of:
        for line in outlist:
            of.write(line)


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


try:
    import subprocess as sp
    import logging
    from blastparser import BlastParser
except ModuleNotFoundError:
    from ncOrtho.blastparser import BlastParser


def perform_reblast(sequence, refblast, cpu, outdir, candidate, mirna_data, dust, msl, cleanup, checkCoorthref):
    logger = logging.getLogger('ncortho')
    logger.setLevel(level=logging.DEBUG)

    blast_command = (
        f'blastn -task blastn -db {refblast} -num_threads {cpu} -dust {dust} '
        f'-outfmt "6 sseqid sstart send sstrand bitscore"'
    )
    blast_call = sp.Popen(
        blast_command, shell=True, stdin=sp.PIPE,
        stdout=sp.PIPE, stderr=sp.PIPE, encoding='utf8'
    )
    # run BLAST
    res, err = blast_call.communicate(sequence)
    if err:
        logger.warning(err)
        return ''
        # continue
    if not cleanup:
        blast_output = f'{outdir}/reBLAST_{candidate}.out'
        with open(blast_output, 'w') as bh:
            for line in res:
                bh.write(line)
    # parse BLASTn results
    blast_output = [line.split() for line in res.split('\n')]
    if not blast_output[0]:
        logger.info('No re-BLAST hits')
        return ''

    # BlastParser will read the temp_fasta file
    bp = BlastParser(mirna_data, blast_output, msl)
    # the parse_blast_output function will return True if the candidate is accepted
    if bp.evaluate_besthit():
        logger.info('Found best hit')
        return sequence

    elif checkCoorthref:
        logger.info('Best hit differs from reference sequence! Doing further checks\n')
        # coorth_out = '{0}/{1}_coorth'.format(outdir, candidate)
        if bp.check_coortholog_ref(sequence, outdir):
            return sequence
    else:
        logger.info('Best hit does not overlap with miRNA location')
        return ''


import pyfaidx
import subprocess as sp
import os
import logging


def heuristic_search(blastdb, rna, query, heuristic, cpu, out):

    def candidate_blast(db, c, evalue, seq, task):
        # extract candidate regions
        blast_command = (
            'blastn -task {0} -db {1} '
            '-num_threads {2} -evalue {3} '
            '-outfmt "6 sseqid sstart send sstrand length"'.format(task, db, c, evalue)
        )
        blast_call = sp.Popen(
            blast_command, shell=True, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE, encoding='utf8'
        )
        res, err = blast_call.communicate(seq)
        if err:
            raise sp.SubprocessError(err)

        result = list(filter(None, res.split('\n')))
        hit_list = [line.split() for line in result if line and float(line.split()[-1]) >= blast_len_cut]

        return hit_list

    def extract_candidate_regions(candidate_list, query, extraregion=1000):
        hit_at_start = False
        hit_at_end = False
        genome = pyfaidx.Fasta(query)

        regions = []
        for hit in candidate_list:
            chrom, start, end, strand, length = hit
            strand = strand.replace('plus', '+').replace('minus', '-')
            if strand == '-':
                start, end = end, start
            start = int(start)
            end = int(end)

            if start > extraregion:
                n_start = start - extraregion
            else:
                # hit starts at the beginning of the chromosome
                n_start = 0
                hit_at_start = True
            n_end = end + extraregion
            if n_end > genome[chrom][-1].end:
                # hit is at the end of the chromosome
                n_end = genome[chrom][-1].end
                hit_at_end = True
            if strand == '+':
                sequence = genome[chrom][n_start:n_end].seq
            elif strand == '-':
                sequence = genome[chrom][n_start:n_end].reverse.complement.seq
            else:
                raise ValueError(f'Unknown strand {strand}')
            regions.append(f">{chrom}|{str(start)}|{str(end)}|{strand}|{hit_at_start}|{hit_at_end}\n{sequence}\n")

        return regions


    ################################################################################################
    blast_len_cut = len(rna.seq) * heuristic[2]
    candidate_list = candidate_blast(blastdb, cpu, heuristic[1], rna.seq, 'blastn')
    if not candidate_list:
        return ''

    candidate_regions = extract_candidate_regions(candidate_list, query)
    fasta = f'{out}/candidate_regions_{rna.name}.fa'
    with open(fasta, 'w') as of:
        for line in candidate_regions:
            of.write(line)

    return fasta


#[
# ['(1)', '!', '3.7e-23', '94.0', '0.0', 'NW_020173731.1|9158415|9158472|+|False|False', '1000', '1057', '+', 'cm', 'no', '0.41', '-'],
# ['(2)', '!', '3.7e-23', '94.0', '0.0', 'NW_020173731.1|9274958|9275015|+|False|False', '1000', '1057', '+', 'cm', 'no', '0.41', '-'],
# ['(3)', '!', '8.3e-15', '62.1', '0.0', 'NW_020172457.1|4920040|4920099|+|False|False', '1000', '1059', '+', 'cm', 'no', '0.33', '-']
# ]

def perform_model_search(models, rna, candidate_region_fasta, output, cm_cutoff, cpu, phmm):

    def parse_phmm(tblout):
        phmm_results = []
        with open(tblout) as fh:
            for line in fh:
                if line.startswith('#'):
                    continue
                data = line.strip().split()
                # print(data)
                info = data[0]
                start, end = data[6:8]
                strand, evalue, score = data[11:14]
                phmm_results.append(['a', 'b', evalue, score, 'c', info, start, end, strand, 'd', 'e', 'f', 'g'])
        return phmm_results

    def parse_cm(res):
        cm_results = []
        for line in res.split('\n'):
            # print(line)
            if line.startswith('  ('):
                if 'No hits detected that satisfy reporting thresholds' in line:
                    return []
                else:
                    data = line.strip().split()
                    cm_results.append(data)
        return cm_results

    ################################################################################
    cut_off = rna.bit * cm_cutoff

    if phmm:
        tblout = f'{output}/{rna.name}_phmm.out'
        call = sp.Popen(
            f'nhmmer --cpu {cpu} --tblout {tblout} -T {cut_off} --incT {cut_off} {models}/{rna.name}.phmm {candidate_region_fasta}',
            shell=True, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE, encoding='utf8'
        )
        res, err = call.communicate(rna.seq)
        if err:
            raise sp.SubprocessError(err)
        results = parse_phmm(tblout)
        os.remove(tblout)
    else:
        call = sp.Popen(
            f'cmsearch --cpu {cpu} --noali -T {cut_off} --incT {cut_off} {models}/{rna.name}.cm {candidate_region_fasta}',
            shell=True, stdout=sp.PIPE, stderr=sp.PIPE, encoding='utf8'
        )
        res, err = call.communicate(rna.seq)
        if err:
            raise sp.SubprocessError(err)
        results = parse_cm(res)

    return results


def parse_cmsearch(cm_results):
    logger = logging.getLogger('ncortho')
    logger.setLevel(level=logging.DEBUG)
    return_data = []
    for data in cm_results:
        score = float(data[3])
        chrom = data[5]
        start, end = map(int, data[6:8])
        strand = data[8]
        return_data.append([chrom, start, end, strand, score])
    return return_data


def parse_cmsearch_for_heuristic(cm_results, extraregion=1000):
    logger = logging.getLogger('ncortho')
    logger.setLevel(level=logging.DEBUG)
    return_data = []
    for data in cm_results:
        # parse CMsearch output of candidate regions to acutal genomic regions
        blast_chrom = data[5].split('|')[0]
        blast_start, blast_end = map(int, data[5].split('|')[1:3])
        hit_at_start, hit_at_end = data[5].split('|')[4:6]
        blast_strand = data[5].split('|')[3]
        cm_start, cm_end = map(int, data[6:8])
        # cm_start, cm_end = map(int, data[5:7])
        cm_strand = data[8]
        if cm_strand == '-':
            # reverse hits should not be possible since blasthits were extracted
            # as reverse complement if they were on the minus strand
            # logger.info('Unexpected hit of CMsearch')
            continue
        if hit_at_start == 'True':
            # print(f'# cmsearch hit for {rna_id} at the beginning of a chromosome in the query species')
            hit_start = cm_start
            hit_end = cm_end
        elif hit_at_end == 'True':
            # print(f'# cmsearch hit for {rna_id} at the end of a chromosome in the query species')
            hit_start = blast_start + (cm_start - extraregion) - 1
            hit_end = hit_start + (cm_end - cm_start) - 1
        else:
            hit_start = blast_start + (cm_start - extraregion) - 1
            hit_end = hit_start + (cm_end - extraregion) - 1
        # fill output variable
        return_data.append([blast_chrom, hit_start, hit_end, blast_strand, float(data[3])])
    return return_data


def remove_duplicates_apply_cutoffs(cms, cmc, lc, rna):
    """
    Parse the output of cmsearch while eliminating duplicates and filtering
    entries according to the defined cutoff.

    Parameters
    ----------
    cms     :   List with CMsearch hits [chrom, start, end, strand, score]
    cmc     :   Cutoff to decide which candidate hits should be included for the reverse BLAST search
    lc      :   Length cutoff
    rna     :   rna class

    Returns
    -------
    hits_dict : Dictionary containing hits of the cmsearch

    """
    # Output
    hits_dict = {}
    # Required for finding duplicates, stores hits per chromosome
    chromo_dict = {}

    for candidate_nr, hit in enumerate(cms, 1):
        # print(hit)
        h_chrom, h_start, h_end, h_strand, h_score = hit
        # blastparser expects the start to be the smaller number, will extract reverse complement if on - strand
        if h_start > h_end:
            h_start, h_end = h_end, h_start
        length = h_end - h_start

        if length <= length * lc:
            continue
        if h_score <= rna.bit * cmc:
            continue

        candidate = f'{rna.name}_c{candidate_nr}'
        data = candidate, h_chrom, h_start, h_end, h_strand, h_score
        hits_dict[candidate] = data

        # Store the hits that satisfy the bit score cutoff to filter
        # duplicates.
        if h_chrom in chromo_dict:
            chromo_dict[h_chrom] = []
        chromo_dict[h_chrom] = [data]
    # print(hits_dict)

    # Loop over the candidate hits to eliminate duplicates.
    for chromo in chromo_dict:
        nrhits = len(chromo_dict[chromo])
        if nrhits > 1:
            for hitnr in range(nrhits):
                start = int(chromo_dict[chromo][hitnr][2])
                stop = int(chromo_dict[chromo][hitnr][3])
                strand = chromo_dict[chromo][hitnr][4]
                score = float(chromo_dict[chromo][hitnr][5])
                for chitnr in range(hitnr +1, nrhits):
                    if strand != chromo_dict[chromo][chitnr][4]:
                        cstart = int(chromo_dict[chromo][chitnr][2])
                        cstop = int(chromo_dict[chromo][chitnr][3])
                        cscore = float(chromo_dict[chromo][chitnr][5])
                        # Test if the two hits from opposite strands
                        # overlap, which means one of them is
                        # (probably) a false positive.
                        # Out of two conflicting hits, the one with the
                        # highest cmsearch bit score is retained.
                        if (
                                start in range(cstart, cstop +1)
                                or stop in range(cstart, cstop +1)
                                or cstart in range(start, stop +1)
                                or cstop in range(start, stop +1)
                        ):
                            if score > cscore:
                                del hits_dict[chromo_dict[chromo][chitnr][0]]

                            else:
                                del hits_dict[chromo_dict[chromo][hitnr][0]]

    return hits_dict


def model_search(rna, cm_cutoff, cpu, msl, models, query, blastdb, out, cleanup, heuristic_col, phmm):
    """
    Parameters
    ----------
    rna       :   rna object
    cm_cutoff   :   Minimum bit score of the CMsearch hit relative to the
                    bit score that results from searching in the reference genome with the same model
    cpu         :   Number of cores to use
    msl         :   Minimum length of CMsearch hit relative to the reference pre-rna length
    models      :   Path to the directory that contains the CMs
    query       :   Path to the query genome
    blastdb     :   Optional Path to a BLASTdb of the query genome
    out         :   Path to the output directory
    cleanup     :   Delete intermediate files True/False
    heuristic   :   Should heuristic mode be used and if yes, set parameters (True/False, evalue, minlength)

    Returns
    -------
    cm_results : Dictionary containing hits of the cmsearch

    """
    # print(os.path.isfile(f'{models}/{rna.name}.phmm'))
    if (
            not phmm and not os.path.isfile(f'{models}/{rna.name}.cm') or
            phmm and not os.path.isfile(f'{models}/{rna.name}.phmm')
    ):
        return False, 'No model found'
    heuristic, heuristic_evalue, heuristic_length, sensitive_heuristic = heuristic_col
    if heuristic:
        candidate_region_fasta = heuristic_search(blastdb, rna, query, heuristic_col, cpu, out)

        if not candidate_region_fasta and sensitive_heuristic:
            # No candidate regions above length cutoff found with BLASTn. Trying again without heuristic.
            cm_results = perform_model_search(models, rna, query, out, cm_cutoff, cpu, phmm)
        elif not candidate_region_fasta:
            return False, 'No candidate region found with BLASTn'
        else:
            cm_results = perform_model_search(models, rna, candidate_region_fasta, out, cm_cutoff, cpu, phmm)
            cm_results = parse_cmsearch_for_heuristic(cm_results)

        if cleanup:
            os.remove(candidate_region_fasta)
    else:
        cm_results = perform_model_search(models, rna, query, out, cm_cutoff, cpu, phmm)
        cm_results = parse_cmsearch(cm_results)

    if not cm_results:
        return False, 'No CMsearch results in candidate regions or query genome'

    parsed_cm_results = remove_duplicates_apply_cutoffs(cm_results, cm_cutoff, msl, rna)
    if not parsed_cm_results:
        return False, 'No CMsearch results above threshold'
    else:
        return parsed_cm_results, 'Sucess'













import subprocess as sp
import os
import shutil

from ncOrtho.utils import check_blastdb
from ncOrtho.utils import make_blastndb


def vprint(s, verbose):
    if verbose:
        print(s, flush=True)


def make_alignment(out, mirna, cpu, core, rcoffee):
    alignment = '{}/{}.aln'.format(out, mirna)
    stockholm = '{}/{}.sto'.format(out, mirna)

    # Call T-Coffee for the sequence alignment.
    # print('Building T-Coffee alignment.')
    if rcoffee == 'yes':
        tc_cmd_1 = (
            f't_coffee -quiet -multi_core={cpu} -mode=rcoffee -in {core} -output=clustalw_aln -outfile={alignment}'
        )
    else:
        tc_cmd_1 = (
            f't_coffee -quiet -multi_core={cpu} -in {core} -output=clustalw_aln -outfile={alignment}'
        )
    sp.run(tc_cmd_1, shell=True, capture_output=True)

    # Extend the sequence-based alignment by structural information.
    # Create Stockholm alignment.
    # print('Adding secondary structure to Stockholm format.')
    if rcoffee == 'yes':
        tc_cmd_2 = (
            f't_coffee -other_pg seq_reformat -in {alignment} -action +add_alifold -output stockholm_aln -out {stockholm}'
        )
    else:
        tc_cmd_2 = (
            f't_coffee -other_pg seq_reformat -in {alignment} -output stockholm_aln -out {stockholm}'
        )
    sp.run(tc_cmd_2, shell=True, capture_output=True)
    return stockholm


def maximum_blast_bitscore(mirna, seq, blastdb, c, dust):
    # The miRNA precursors can show a low level of complexity, hence it might be
    # required to deactivate the dust filter for the BLAST search.
    bit_check = (
        f'blastn -num_threads {c} -dust {dust} -task megablast -db {blastdb} -outfmt \"6 bitscore\"'
    )
    # print('# Determining reference bit score..')
    ref_bit_cmd = sp.Popen(
        bit_check, shell=True, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.STDOUT, encoding='utf8'
    )
    ref_results, err = ref_bit_cmd.communicate(seq)
    if not ref_results:
        print(f'WARNING: Reference sequence of {mirna} not found in reference Genome. Setting maximum bitscore to 0', flush=True)
        return 0.0
    try:
        ref_bit_score = float(ref_results.split('\n')[0].split('\t')[0])
    except ValueError:  # BLASTn errors are in output not in error
        print(f'WARNING: Reference sequence of {mirna} not found in reference Genome. Setting maximum bitscore to 0', flush=True)
        return 0.0
    return ref_bit_score


# Perform reciprocal BLAST search and construct Stockholm alignment
def blastsearch(mirna, r_path, o_path, c, dust, v, coffee):
    """

    Parameters
    ----------
    m_path  :   Path to file with miRNA information
    r_path  :   Path to reference genome.
    o_path  :   Output Name
    c       :   Number of Threads for BLAST search.
    dust    :   Use dust filter for BLAST searches (yes/no)

    Returns
    -------
    Path to Alignment of core pre-miRNAs in Stockholm format

    """

    mirid, rawchrom, mstart, mend, mstrand, rawseq = mirna[:6]
    mchr = rawchrom.replace('chr', '')
    preseq = rawseq.replace('U', 'T').replace('-', '')

    miroutdir = f'{o_path}/{mirid}'
    if not os.path.isdir(miroutdir):
        os.mkdir(miroutdir)
    synteny_regs = f'{miroutdir}/synteny_regions_{mirid}.fa'  # this fasta file is created by the the main() script
    os.chdir(miroutdir)

    print(f'# {mirid}', flush=True)
    if not os.path.isfile(synteny_regs):
        print(f'Warning: No synteny regions found for {mirid}. Training with reference miRNA only.', flush=True)
        with open(synteny_regs, 'w') as fastah:
            fastah.write(f'>{mirid}\n{preseq}\n')
        stock = make_alignment(miroutdir, mirid, c, synteny_regs, coffee)
        return stock

    # check if blastdb of reference genome exists
    fname = '.'.join(r_path.split("/")[-1].split('.')[0:-1])
    ref_blastdb = f'{o_path}/refBLASTdb/{fname}'
    if not check_blastdb(ref_blastdb):
        make_blastndb(r_path, ref_blastdb)

    max_bitscore = maximum_blast_bitscore(mirid, preseq, ref_blastdb, c, dust)

    # Find pre-miRNA candidates in syntenic region'
    if not check_blastdb(synteny_regs):
        make_blastndb(synteny_regs, synteny_regs)
    blastn_cmd = (
        'blastn -num_threads {0} -task blastn -dust {1} -db {2} -outfmt \"6 '
        'sseqid evalue bitscore sseq\"'.format(c, dust, synteny_regs)
    )
    blastn = sp.Popen(
        blastn_cmd, shell=True, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE, encoding='utf8'
    )
    ortholog_candidates, err = blastn.communicate(preseq)
    if err:
        raise sp.SubprocessError(err)

    # Re-BLAST
    outputcol = {}
    seq_check = set()
    ortholog_candidates_list = ortholog_candidates.split('\n')
    vprint(f'Number of ortholog candidates: {len(ortholog_candidates_list)}', v)
    for hit in ortholog_candidates_list:
        if not hit:
            continue
        core_region, eval, bitscore, sseq = hit.strip().split()
        if float(bitscore) <= max_bitscore * 0.5:
            continue

        degap_seq = sseq.replace('-', '')  # Eliminate gaps in BLAST output
        reblastn_cmd = (
            f'blastn -num_threads {c} -task blastn -dust {dust} -db {ref_blastdb} -outfmt \"6'
            ' sseqid sstart send\"'
        )
        reblastn = sp.Popen(
            reblastn_cmd, shell=True, stdin=sp.PIPE,
            stdout=sp.PIPE, stderr=sp.STDOUT, encoding='utf8'
        )
        reresults, reerr = reblastn.communicate(degap_seq)
        for reblast_res in reresults.split('\n'):
            if not reblast_res:
                continue
            refchrom, refstart, refend = reblast_res.split()

            if refchrom != mchr:
                vprint('Hit on chromosome {}, expected {}'.format(refchrom, mchr), v)
                continue
            if (refstart <= mstart <= refend) or (refstart <= mend <= refend):  # first within second
                vprint('Reciprocity fulfilled.', v)
            elif (mstart <= refstart <= mend) or (mstart <= refend <= mend):  # second within first
                vprint('Reciprocity fulfilled.', v)
            else:
                vprint('Reciprocity unfulfilled.', v)
                continue
            if core_region not in seq_check:  # only train with best hit per syntenic region
                outputcol[core_region] = degap_seq
                seq_check.add(core_region)

    corefile = f'{miroutdir}/core_orthologs_{mirid}.fa'
    with open(corefile, 'w') as outfile:
        outfile.write(f'>reference\n{preseq}\n')
        if outputcol:
            for synteny_region, sequence in outputcol.items():
                outfile.write(f'>{synteny_region}\n{sequence}\n')
        else:
            print(f'Warning: No core orthologs found for {mirid}. Training with reference miRNA only.', flush=True)
    stock = make_alignment(miroutdir, mirid, c, corefile, coffee)
    return stock



import sys
import re
import yaml


def vprint(s, verbose):
    if verbose:
        print(s, flush=True)

# Parse an Ensembl GTF file to store the coordinates for each protein-coding gene in a
# dictionary
def gtf_parser(gtf):
    species_name = gtf.split('/')[-1].split('.')[0]
    chr_dict = {}
    chromo = ''

    # with open(inpath) as infile, open(outpath, 'wb') as outfile:
    with open(gtf) as infile:
        for line in infile:
            if (
                    not line.startswith('#')
                    and line.split()[2] == 'gene'
                    and line.split('gene_biotype')[1].split('\"')[1]
                    == 'protein_coding'
            ):
                linedata = line.strip().split('\t')
                contig = linedata[0]
                geneid = linedata[-1].split('\"')[1]
                start = int(linedata[3])
                end = int(linedata[4])
                strand = linedata[6]
                if contig != chromo:
                    i = 1
                    chromo = contig
                chr_dict[geneid] = (contig, i)
                try:
                    chr_dict[contig][i] = (geneid, start, end, strand)
                except:
                    chr_dict[contig] = {i: (geneid, start, end, strand)}
                i += 1
    return chr_dict


def gff_parser(gff, id_type):
    """

    Parameters
    ----------
    gff :   Path to a RefSeq gff3 file
    id  :   type of identifier in the gff3 file that is to be compared against the IDs in the pairwise ortholgs file

    Returns
    -------
    Dictionary containing the location of each gene on a chromosome
    {'gene_id': ('chromosome': position), 'chromosome': {position: ('gene_id', start, end, strand)} }

    """
    chr_dict = {}
    chromo = ''
    # with open(inpath) as infile, open(outpath, 'wb') as outfile:
    with open(gff) as infile:
        for line in infile:
            if line.startswith('#'):
                continue
            if 'gene_biotype=protein_coding' not in line:
                continue
            linedata = line.strip().split('\t')
            if linedata[2] != 'gene':
                continue

            if id_type == 'GeneID':
                geneid = re.split('[;,]', linedata[-1].split(f'{id_type}:')[1])[0]
            elif id_type in ['ID', 'Name', 'CDS']:
                geneid = linedata[-1].split(f'{id_type}=')[1].split(';')[0]
                if id_type in ['ID', 'CDS']:
                    geneid = '-'.join(geneid.split('-')[1:])
            else:
                raise ValueError('Unknown identifier type "{}"'.format(id_type))

            contig = linedata[0]
            start = int(linedata[3])
            end = int(linedata[4])
            strand = linedata[6]
            if contig != chromo:  # reset counter
                i = 1
                chromo = contig
            chr_dict[geneid] = (contig, i)
            if contig not in chr_dict:
                chr_dict[contig] = {}
            chr_dict[contig][i] = (geneid, start, end, strand)

            i += 1
    return chr_dict


def table_parser(path):
    chromo = ''
    chr_dict = {}
    with open(path) as fh:
        for line in fh:
            if line.startswith('#'):
                continue
            contig, start, end, strand, geneid = line.strip().split('\t')
            if contig != chromo:
                i = 1
                chromo = contig
            chr_dict[geneid] = (contig, i)
            try:
                chr_dict[contig][i] = (geneid, int(start), int(end), strand)
            except KeyError:
                chr_dict[contig] = {i: (geneid, int(start), int(end), strand)}
            i += 1
    return chr_dict


def parse_annotation(path, idtyp):
    ft = path.split('.')[-1]
    if ft == 'gtf':
        rd = gtf_parser(path)
    elif ft in ['gff3', 'gff']:
        rd = gff_parser(path, idtyp)
    elif ft in ['tsv', 'txt']:
        rd = table_parser(path)
    else:
        raise ValueError(f'File type "{idtyp}" not valid as reference annotation. Expecting .gff3, .gff or .gtf')
    return rd


###############################################################################

def parse_yaml(path):
    p_dict = {}  # {<name>: {'orthologs': <path>, 'genome': <path>, 'annotation': <path>}}
    paths = []
    refdict = {}
    with open(path, 'r') as param_handle:
        params = yaml.load_all(param_handle, Loader=yaml.FullLoader)
        for entry in params:
            in_type = entry.pop('type')
            if in_type == 'reference':
                refdict['annotation'] = entry['annotation']
                refdict['genome'] = entry['genome']
            elif in_type == 'core':
                species = entry.pop('name')
                p_dict[species] = entry
            paths.extend([entry['annotation'], entry['genome']])
        return p_dict, refdict, paths


def read_pairwise_orthologs(pathdict):
    od = {}
    for taxon, pd in pathdict.items():
        taxd = {}
        orthofile = pd['orthologs']
        with open(orthofile) as omafile:
            for line in omafile:
                if line.startswith('#'):
                    continue
                ref, core = line.strip().split()[:2]
                if not ref in taxd:
                    taxd[ref] = []
                taxd[ref].append(core)
        od[taxon] = taxd
    return od


def gene_from_cds(gff):
    """
    returns:    d = {'WP_23898234': 'CDK'}
    """
    # cds2parent = {}
    cds_mrna_map = {}

    cds2gene = {}
    orphan = []
    with open(gff) as fh:
        for line in fh:
            if line.startswith('#'):
                continue
            data = line.strip().split('\t')
            if data[2] != 'CDS':
                continue
            cdsid = data[-1].split(';')[0].split('.')[0].replace('ID=cds-', '').replace('ID=id-', '')

            parent = data[-1].split('Parent=')[1].split(';')[0].split('.')[0]
            if parent.startswith('gene-') or parent.startswith('id-'):
                cds2gene[cdsid] = parent.split(';')[0].replace('gene-', '').replace('id-', '')
            else:
                cds_mrna_map[cdsid] = parent
                cds_mrna_map[parent] = cdsid

    with open(gff) as fh:
        for line in fh:
            if line.startswith('#'):
                continue
            data = line.strip().split('\t')
            if data[2] != 'mRNA':
                continue
            mrnaid = data[-1].split('ID=')[1].split('.')[0]
            parent = data[-1].split('Parent=')[1].split('.')[0]
            if mrnaid not in cds_mrna_map:
                continue
            cdsid = cds_mrna_map[mrnaid]
            if parent.startswith('gene-'):
                cds2gene[cdsid] = parent.split(';')[0].replace('gene-', '')
            else:
                orphan.append(parent)
    return cds2gene


def pairwise_orthologs_from_cds(cd, refanno):
    """
    returns    d = {corespec: {refprot: [ortho1, ortho2, ...], ...}, ...}
    """
    ref_cds2gene = gene_from_cds(refanno)
    od = {}
    for cspec in cd:
        cds2gene = gene_from_cds(cd[cspec]['annotation'])

        od[cspec] = {}
        orthopath = cd[cspec]['orthologs']
        with open(orthopath) as fh:
            for line in fh:
                cds_refprot, cds_orthoprot = line.strip().split()
                try:
                    refprot = ref_cds2gene[cds_refprot]
                except KeyError as e:
                    # print('reference', e)
                    # IDs of "longest proteins" that were removed in later versions of the RefSeq annotation.
                    # Discrepency possibly due to versioning problems
                    continue
                try:
                    orthoprot = cds2gene[cds_orthoprot]
                except KeyError as e:
                    print(cspec, e)
                    continue

                if not refprot in od[cspec]:
                    od[cspec][refprot] = []
                od[cspec][refprot].append(orthoprot)
    return od


def read_mirnas(path):
    with open(path) as mirfile:
        mirnas = [
            line.split('\t') for line in mirfile.readlines()
            if not line.startswith('#')
        ]
    # print('Done')
    return mirnas


def mirna_position(mirlist):
    mirid, rawchrom, start, end, strand = mirlist[:5]
    if 'chr' in rawchrom:
        chromo = rawchrom.split('chr')[1]
    else:
        chromo = rawchrom
    return mirid, chromo, int(start), int(end), strand









def vprint(s, verbose):
    if verbose:
        print(s, flush=True)


def no_synteny_possible(mirna, chrom, start, end, reference):
    firstgene_info = reference[chrom][1]
    firstgene_start = int(firstgene_info[1])
    lastgene_info = reference[chrom][len(reference[chrom])]
    lastgene_end = int(lastgene_info[2])
    # case 1): there is no protein-coding gene on the same contig as the miRNA,
    # so there can be no neighbors (should only occur in highly fragmented
    # assemblies)
    if chrom not in reference.keys():
        print(
            f'WARNING: No protein-coding genes found on contig "{chrom}". '
            'Make sure that the contig identifiers of the miRNA input file '
            'match the ones in the reference annotation file', flush=True
        )
        return True
    # case 2): miRNA is located left of the first gene and hence has no left
    # neighbor, the first gene is therefore by default the right neighbor

    elif end < firstgene_start:
        print(
            f'There is no left neighbor of {mirna}, since it is located at the start of contig {chrom}\n'
            f'{firstgene_info[0]} is the right neighbor of {mirna}', flush=True
        )
        return True
    # case 3): miRNA is located right to the last gene, so the last gene is the
    # left neighbor and there cannot be a right neighbor
    elif start > lastgene_end:
        print(
            f'There is no right neighbor of {mirna}, since it is located at the end of contig {chrom}\n'
            f'{lastgene_info[0]} is the left neighbor of {mirna}', flush=True)
        return True
    else:
        return False


# def ortho_search(r_gene, core_dict, v):
#     """
#     Try to find the ortholog for a given reference gene in a core set species
#
#     Parameters
#     ----------
#     r_gene:     Gene of Reference Species
#     ortho_dict: {'corespecies': {r_gene: [ortho1], ...}, ...}
#
#     Returns:    {'corespecies': [coreortho1], ...}
#     -------
#
#     """
#     coreorthologs_per_species = {}
#     for core_taxon, orthodict in core_dict.items():
#         if r_gene in orthodict:
#             core_orthologs = orthodict[r_gene]
#             coreorthologs_per_species[core_taxon] = core_orthologs
#             vprint(f'{" ".join(core_orthologs)} identified as ortholog(s) to {r_gene} in {core_taxon}', v)
#         else:
#             vprint(f'No ortholog found for {r_gene} in {core_taxon}', v)
#     return coreorthologs_per_species


def neighbor_search(leftgene, rightgene, core_dict, gene_position, chromdict, no_next_orthos, v):

    def find_ortho(genename, position, chromd, orthologs, next_orthos, typ, v):
        if genename in orthologs:
            core_orthologs = orthologs[genename]
            vprint(f'{" ".join(core_orthologs)} identified as ortholog(s) to {genename} in {core_taxon}', v)
            return core_orthologs
        else:
            for index in range(1, next_orthos):
                if position - index < 1:
                    continue
                if position + index > len(chromd):
                    continue

                if typ == 'left':
                    nextgeneinfo = chromd[position - index]
                else:
                    nextgeneinfo = chromd[position + index]
                nextgene = nextgeneinfo[0]
                if nextgene in orthologs:
                    core_orthologs = orthologs[nextgene]
                    vprint(f'No  ortholog(s) to {genename} in {core_taxon}', v)
                    vprint(f'{" ".join(core_orthologs)} identified as ortholog(s) of {nextgene}, '
                           f'which is in distance of '
                           f'{index} genes to {genename} in {core_taxon}', v)
                    return core_orthologs

    coreorthologs_per_species = {}
    for core_taxon, orthodict in core_dict.items():
        leftorthos = find_ortho(leftgene, gene_position, chromdict, core_dict[core_taxon], no_next_orthos, 'left', v)
        rightorthos = find_ortho(rightgene, gene_position + 1, chromdict, core_dict[core_taxon], no_next_orthos, 'right', v)
        if leftorthos and rightorthos:
            coreorthologs_per_species[core_taxon] = (leftorthos, rightorthos)
    return coreorthologs_per_species


def categorize_mirna_position(
        mirna, mirna_chrom, mirna_start, mirna_end, mirna_strand, reference, all_orthologs, no_add_orthos, v
):
    """

    Parameters
    ----------
    reference:      Dictionary of reference species annotation as returned by GFF3-Parser
    all_orthologs:  Dictionary of Orthologs as returned by read_pairwise_orthologs
    no_add_orthos:  Number of additional genes up- or downstream that may also serve as syntenic anchors

    Returns

    -------

    """
    syntenytype = None
    solved = False
    core_orthologs = []
    if no_synteny_possible(mirna, mirna_chrom, mirna_start, mirna_end, reference):
        return solved, syntenytype

    # case 4): miRNA is located either between two genes or overlapping with (an
    # intron of) a gene, either on the same or the opposite strand
    ###############################################################################
    for position, geneinfo in reference[mirna_chrom].items():
        gene, gene_start, gene_end, gene_strand = geneinfo
        if mirna_start >= gene_start and mirna_end <= gene_end:
            # case 4.1): miRNA inside gene
            if mirna_strand == gene_strand:
                syntenytype = 'inside'
                solved = True
                vprint(f'{mirna} is located inside the gene {gene}', v)
            # case 4.2): miRNA opposite of gene
            else:
                syntenytype = 'opposite'
                solved = True
                vprint(f'{mirna} is located opposite of the gene {gene}', v)
            # core_orthologs = ortho_search(gene, all_orthologs, v)
            core_orthologs = neighbor_search(
                gene, gene, all_orthologs, position, reference[mirna_chrom], no_add_orthos, v
            )

        # case 4.3): miRNA between genes
        else:
            if position == len(reference[mirna_chrom]):
                continue
            rightneighborinfo = reference[mirna_chrom][position + 1]
            rightneighbor = rightneighborinfo[0]
            rn_start = rightneighborinfo[1]
            rn_end = rightneighborinfo[2]

            # if gene_end < mirna_start and rn_start > mirna_end:
            if gene_start <= mirna_start and rn_end >= mirna_end:
                syntenytype = 'in-between'
                solved = True
                vprint(f'{gene} is the left neighbor of {mirna}, {rightneighbor} is the right neighbor', v)
                core_orthologs = neighbor_search(
                    gene, rightneighbor, all_orthologs, position, reference[mirna_chrom], no_add_orthos, v
                )
    if solved:
        return syntenytype, core_orthologs
    else:
        vprint(f'Unable to resolve synteny for {mirna}', v)
        return None, None



import glob
import os
import tempfile
import subprocess as sp
import sys
import shutil
import inspect
from collections import Counter
import argparse
from pyfiglet import Figlet
from importlib.metadata import version

"""
Species in spec_to_skip will not be part of multiple sequence alignments
"""


def make_supermatrix(out):
    align_out = '{}/alignments'.format(out)
    tree_out = '{}/supermatrix'.format(out)
    # curr_dir = sys.path[0]
    curr_dir = '/'.join(inspect.getfile(inspect.currentframe()).split('/')[:-1])
    if not os.path.isdir(tree_out):
        os.mkdir(tree_out)
    print('# Creating supermatrix alignment')
    cmd = (
        'perl {}/concat_alignments_dmp.pl -in {} -out supermatrix.aln'
            .format(curr_dir, align_out)
    )
    res = sp.run(cmd, shell=True, capture_output=True)
    if res.returncode != 0:
        print('ERROR:')
        print(res.stderr.decode('utf-8'))
        sys.exit()
    # print('# Done')

    print('# De-gapping alignment')
    cmd = (
        'perl {}/degapper.pl -in {}/supermatrix.aln'.format(curr_dir, out)
    )
    sp.run(cmd, shell=True)
    # move files to output dir
    shutil.move(f'{out}/supermatrix.aln', f'{tree_out}/supermatrix.aln')
    shutil.move(f'{out}/subalignment_positions.txt', f'{tree_out}/subalignment_positions.txt')
    shutil.move(f'{out}/supermatrix.aln.proc', f'{tree_out}/supermatrix.aln.proc')
    # print('# Done')
    # return f'{tree_out}/supermatrix.aln'
    return f'{tree_out}/supermatrix.aln.proc'


def arglistcheck(arg):
    if not arg:
        return ''
    if os.path.isfile(arg):
        with open(arg) as fh:
            arglist = [element.strip() for element in fh.read().split('\n') if element]
    else:
        arglist = [element.strip() for element in arg.split(',') if element]
    return arglist


def parse_matseq(path):
    m2s = {}
    with open(path) as fh:
        for line in fh:
            if line.startswith('#'):
                continue
            mature = line.strip().split('\t')[-1].replace('U', 'T')
            mirid = line.strip().split('\t')[1].split('_')[0]
            if len(mature) < 7:
                raise ValueError('No mature sequence found in "ncrna_file"')
            seed = mature[1:8]
            m2s[mirid] = seed
    return m2s

def main():

    # Parse command-line arguments
    # Define global variables
    parser = argparse.ArgumentParser(
        description='Analzye ncOrtho results. Gives: PhyloProfile input, supermatrix alignment, species tree.'
    )
    parser._action_groups.pop()
    required = parser.add_argument_group('Required Arguments')
    optional = parser.add_argument_group('Optional Arguments')
    required.add_argument(
        '-r', '--results', metavar='<path>', type=str,
        help='Path to ncOrtho output directory that is to be analyzed'
    )
    required.add_argument(
        '-o', '--output', metavar='<path>', type=str,
        help='Path to location where output of analysis should be stored. '
             'Assumes that result files end with "_orthologs.fa"'
    )
    required.add_argument(
        '-m', '--mapping', metavar='<path>', type=str,
        help='Tab seperated mapping file between NCBI taxonomy id and name of the query species '
             'or more generally the name of the directory that contains the ncOrtho results of each species'
             '(e.g "9606\tHomo_sapiens")'
    )
    ##########################################################################################
    optional.add_argument(
        '--skip', metavar='str', type=str, nargs='?', const='', default='',
        help=(
            'Comma separated list or path to newline seperated list of species for which analyses should be skipped'
        )
    )
    optional.add_argument(
        '--include', metavar='str', type=str, nargs='?', const='', default='',
        help=(
            'Comma separated list or path to newline seperated list of species for which analyses should be performed. '
            'Will skip every species not in list.'
        )
    )
    optional.add_argument(
        '--auto_skip', metavar='float', type=float, nargs='?', const=0.0, default=0.0,
        help=(
            'Exclude species for which less than a fraction of the reference '
            'miRNAs were detected from species tree calculation. '
            '(e.g. --auto_skip 0.5 means that only species in which at least 50%% '
            'of the reference miRNAs were found are used for species tree calculation) (Default: Off)'
        )
    )
    optional.add_argument(
        '--iqtree', metavar='str', type=str, nargs='?',
        const='',
        default='',
        help=(
            'Call to iqtree. Do not change curly bracket notation '
            '(Default: iqtree -s {} -bb 1000 -alrt 1000 -nt AUTO -redo -pre {}/{})'
        )
    )
    optional.add_argument(
        '--mirnas', metavar='<path>', type=str, nargs='?', const='', default='',
        help=(
            'Comma separated list or path to newline seperated file with miRNAs that should be used for reconstruction'
        )
    )
    # mirna data
    optional.add_argument(
        '--ncrna_file', metavar='<path>', nargs='?', const='', default='',
        help='Path to tab separated file of reference miRNAs information, as used in ncCreate and ncSearch. '
             'Uses mature sequence column to identify seed conservation.'
    )

    # print header
    custom_fig = Figlet(font='stop')
    print(custom_fig.renderText('ncOrtho')[:-3], flush=True)
    v = version('ncOrtho')
    print(f'Version: {v}\n', flush=True)
    ##########################################################################################
    # Parse arguments
    ##########################################################################################
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    else:
        args = parser.parse_args()

    res_dict = args.results
    taxmap = args.mapping
    outdir = args.output

    spec_to_skip = arglistcheck(args.skip)
    spec_include = arglistcheck(args.include)

    overlap = set(spec_to_skip).intersection(set(spec_include))
    if overlap:
        raise ValueError(f'"{overlap}" present in species to skip as well as species to include')


    mirlist = arglistcheck(args.mirnas)

    auto_skip = args.auto_skip

    if args.ncrna_file:
        mirid2seed = parse_matseq(args.ncrna_file)
    else:
        mirid2seed = {}
    ##########################################################################################
    # Argument checks
    ##########################################################################################
    if not os.path.isdir(outdir):
        os.mkdir(outdir)
    if auto_skip < 0 or auto_skip > 1:
        raise ValueError(f'--auto_skip must be a value between 0 and 1, not {auto_skip}')

    ###############################################################################################
    # Main
    ###############################################################################################
    # find results
    ortholog_files = glob.glob(f'{res_dict}/*/*_orthologs.fa')
    if not ortholog_files:
        ortholog_files = glob.glob(f'{res_dict}/*.fa')
    # read mapping file
    name_2_taxid = {}
    with open(taxmap, 'r') as mp:
        for line in mp:
            taxid, name = line.strip().split('\t')
            name_2_taxid[name] = taxid

    ortho_dict = {}  # ortho_dict = {mirna: {species: seq}}
    # read all results into one dictionary and write phyloprofile input
    pp_in = f'{outdir}/PhyloProfile.long'
    spec_list = []
    with open(pp_in, 'w') as pp:
        pp.write('geneID\tncbiID\torthoID\n')
        for file in ortholog_files:
            with open(file, 'r') as fh:
                for line in fh:
                    if line.startswith('>'):
                        spec, mirna = line.strip().replace('>', '').split('|')[:2]
                    else:
                        seq = line.strip()
                        # write phyloprofile input
                        taxstr = f'ncbi{name_2_taxid[spec]}'
                        group_str = mirna.split('_')[0]
                        if mirid2seed:
                            seed = mirid2seed[group_str]
                            if seed in seq:
                                seedcon = 1
                            else:
                                seedcon = 0
                            pp.write(f'{group_str}\t{taxstr}\t{mirna}\t{seedcon}\n')
                        else:
                            pp.write(f'{group_str}\t{taxstr}\t{mirna}\n')
                        # initialize dictionary
                        if group_str not in ortho_dict:
                            ortho_dict[group_str] = {}
                        # only save representative (first) hit
                        if spec not in ortho_dict[group_str]:
                            ortho_dict[group_str][spec] = seq
                            spec_list.append(spec)
    print('# Finished writing PhyloProfile input')

    # if miRNAfile option is chosen, prune dictionary
    if mirlist:
        tmpdict = {}
        for mirnaid, value in ortho_dict.items():
            if mirnaid in mirlist:
                tmpdict[mirnaid] = value
        del ortho_dict
        ortho_dict = tmpdict

    # count species and skip species with few orthologs before aligning
    if auto_skip:
        spec_counter = Counter(spec_list)
        max_spec_count = spec_counter.most_common(1)[0][1]
        count_cutoff = max_spec_count * auto_skip
        print(f'# Only species with at least {round(count_cutoff)} identified orthologs are used for tree calculation')
        for species, count in dict(spec_counter).items():
            if count < count_cutoff:
                spec_to_skip.append(species)
        print('# Skipping these species:')
        for spsk in spec_to_skip:
            print(spsk)

    # if no species list is given to include, all species are part of specinclude
    if not spec_include:
        spec_include = list(ortho_dict.keys())

    # make alignments
    print('# Starting alignments')
    align_out = f'{outdir}/alignments'
    if not os.path.isdir(align_out):
        os.mkdir(align_out)
    for mirna in ortho_dict:
        # tmplist = []
        if len(ortho_dict[mirna]) < 2:  # need at least 2 sequences for alignment
            with open(f'{align_out}/{mirna}.aln', 'w') as of:
                for spec, seq in ortho_dict[mirna].items():
                    of.write(f'>{spec}\n{seq}\n')
            continue
        with tempfile.NamedTemporaryFile(mode='w+') as fp:
            for spec, seq in ortho_dict[mirna].items():
                if spec_to_skip:
                    if (
                            spec in spec_to_skip
                            and spec not in spec_include
                    ):
                        continue
                fp.write(f'>{spec}\n{seq}\n')
                # tmplist.append(f'>{spec}\n{seq}\n')
            fp.seek(0)
            aln_cmd = f'muscle -align {fp.name} -output {align_out}/{mirna}.aln'
            res = sp.run(aln_cmd, shell=True, capture_output=True)
            if res.returncode != 0:
                # with open(f'{outdir}/faulty_{mirna}.fa', 'w') as of:
                    # for line in tmplist:
                    #     of.write(line)
                print(res.stderr.decode('utf8'))
                sys.exit()

    sys.stdout.flush()
    superm_path = make_supermatrix(outdir)
    print('# Starting tree calculation')
    tree_out = f'{outdir}/species_tree'
    if not os.path.isdir(tree_out):
        os.mkdir(tree_out)

    if not args.iqtree:
        tree_cmd = f'iqtree -s {superm_path} -bb 1000 -alrt 1000 -nt AUTO -redo -pre {tree_out}/species_tree'
    else:
        tree_cmd = args.iqtree
    res = sp.run(tree_cmd, shell=True, capture_output=True)
    print(res.stdout.decode('utf-8'))
    if res.returncode != 0:
        print(res.stderr.decode('utf-8'))

    # write tree that can be used as optional phyloprofile input
    treepath = f'{tree_out}/species_tree.contree'
    pptree = f'{outdir}/PhyloProfile_tree.contree'
    with open(treepath, 'r') as th:
        treestr = th.read().strip()
        for name in name_2_taxid:
            ppid = f'ncbi{name_2_taxid[name]}'
            treestr = treestr.replace(name, ppid)
    with open(pptree, 'w') as otfh:
        otfh.write(treestr)


if __name__ == "__main__":
    main()



"""
ncOrtho - Targeted ortholog search for miRNAs
Copyright (C) 2021 Felix Langschied

ncOrtho is a free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

ncOrtho is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ncOrtho.  If not, see <http://www.gnu.org/licenses/>.
"""


# Modules import
import argparse
import multiprocessing as mp
import os
import sys
from time import time
from tqdm import tqdm
import logging
from pyfiglet import Figlet
from importlib.metadata import version

# Internal ncOrtho modules
try:
    from blastparser import ReBlastParser
    from genparser import GenomeParser
    from cmsearch import model_search
    from utils import check_blastdb, make_blastndb, write_output, str2bool
    from reblast import perform_reblast
    from rna_object import rna_maker
except ModuleNotFoundError:
    from ncOrtho.blastparser import ReBlastParser
    from ncOrtho.genparser import GenomeParser
    from ncOrtho.cmsearch import model_search
    from ncOrtho.utils import check_blastdb, make_blastndb, write_output, str2bool
    from ncOrtho.reblast import perform_reblast
    from ncOrtho.rna_object import rna_maker


# Main function
def main():
    ##########################################################################################
    # Command line arguments
    ##########################################################################################
    
    # Parse command-line arguments
    # Define global variables
    parser = argparse.ArgumentParser(
        description='Find orthologs of reference miRNAs in the genome of a query species.'
    )
    parser._action_groups.pop()
    required = parser.add_argument_group('Required Arguments')
    optional = parser.add_argument_group('Optional Arguments')
    # covariance models folder
    required.add_argument(
        '-m', '--models', metavar='<path>', type=str, required=True,
        help='Path to directory containing covariance models (.cm)'
    )
    # mirna data
    required.add_argument(
        '-n', '--ncrna', metavar='<path>', type=str, required=True,
        help='Path to Tab separated file with information about the reference miRNAs'
    )
    # output folder
    required.add_argument(
        '-o', '--output', metavar='<path>', type=str, required=True,
        help='Path to the output directory'
    )
    # query genome
    required.add_argument(
        '-q', '--query', metavar='<.fa>', type=str, required=True,
        help='Path to query genome in FASTA format'
    )
    # reference genome
    required.add_argument(
        '-r', '--reference', metavar='<.fa>', type=str, required=True,
        help='Path to reference genome in FASTA format'
    )
    ##########################################################################
    # Optional Arguments
    ##########################################################################
    # query_name
    optional.add_argument(
        '--queryname', metavar='str', type=str, nargs='?', const='', default='',
        help=(
            'Name for the output directory (RECOMMENDED) '
        )
    )
    # cpu, use maximum number of available cpus unless specified otherwise
    optional.add_argument(
        '--cpu', metavar='int', type=int,
        help='Number of CPU cores to use (Default: all available)', nargs='?',
        const=mp.cpu_count(), default=mp.cpu_count()
    )
    # bit score cutoff for cmsearch hits
    optional.add_argument(
        '--cm_cutoff', metavar='float', type=float,
        help='CMsearch bit score cutoff, given as ratio of the CMsearch bitscore '
             'of the CM against the refernce species (Default: 0.5)', nargs='?', const=0.5, default=0.5
    )
    # length filter to prevent short hits
    optional.add_argument(
        '--minlength', metavar='float', type=float,
        help='CMsearch hit in the query species must have at '
             'least the length of this value times the length of the refernce pre-miRNA (Default: 0.7)',
        nargs='?', const=0.7, default=0.7
    )
    optional.add_argument(
        '--heuristic', type=str2bool, metavar='True/False', nargs='?', const=True, default=True,
        help=(
            'Perform a BLAST search of the reference miRNA in the query genome to identify '
            'candidate regions for the CMsearch. Majorly improves speed. (Default: True)'
        )
    )
    optional.add_argument(
        '--heur_blast_evalue', type=float, metavar='float', nargs='?', const=0.5, default=0.5,
        help=(
            'Evalue filter for the BLASTn search that '
            'determines candidate regions for the CMsearch when running ncOrtho in heuristic mode. '
            '(Default: 0.5) (Set to 10 to turn off)'
        )
    )
    optional.add_argument(
        '--heur_blast_length', type=float, metavar='float', nargs='?', const=0.5, default=0.5,
        help=(
            'Length cutoff for BLASTN search with which candidate regions for the CMsearch are identified.'
            'Cutoff is given as ratio of the reference pre-miRNA length '
            '(Default: 0.5) (Set to 0 to turn off)'
        )
    )
    optional.add_argument(
        '--sensitive_heuristic', type=float, metavar='True/False', nargs='?', const=False, default=False,
        help=(
            'If no candidate region is found via BLASTn, search with CM in full query genome (Default: False)'
        )
    )
    optional.add_argument(
        '--cleanup', type=str2bool, metavar='True/False', nargs='?', const=True, default=True,
        help=(
            'Cleanup temporary files (Default: True)'
        )
    )
    optional.add_argument(
        '--phmm', type=str2bool, metavar='True/False', nargs='?', const=False, default=False,
        help=(
            'Use pHMM instead of CM for ortholog search (Default: False)'
        )
    )
    optional.add_argument(
        '--refblast', type=str, metavar='<path>', nargs='?', const='', default='',
        help=(
            'Path to BLASTdb of the reference species'
        )
    )
    optional.add_argument(
        '--queryblast', type=str, metavar='<path>', nargs='?', const='', default='',
        help=(
            'Path to BLASTdb of the query species'
        )
    )
    optional.add_argument(
        '--maxcmhits', metavar='int', nargs='?', const=None, default=None,
        help=(
            'Maximum number of cmsearch hits to examine. '
            'Decreases runtime significantly if reference miRNA in genomic repeat region. '
            'Set to empty variable to disable (i.e. --maxcmhits=None, default)'
        )
    )
    # use dust filter?
    optional.add_argument(
        '--dust', metavar='yes/no', type=str,
        help='Use BLASTn dust filter during re-BLAST. Greatly decreases runtime if reference miRNA(s) '
             'are located in repeat regions. '
             'However ncOrtho will also not identify orthologs for these miRNAs (Default: no)',
        nargs='?',
        const='no', default='no'
    )
    # check Co-ortholog-ref
    optional.add_argument(
        '--checkCoorthologsRef', type=str2bool, metavar='True/False', nargs='?', const=False, default=False,
        help=(
            'If the re-blast does not identify the original reference miRNA sequence as best hit,'
            'ncOrtho will check whether the best blast '
            'hit is likely a co-ortholog of the reference miRNA relative to the search taxon. '
        )
    )

    # print header
    custom_fig = Figlet(font='stop')
    print(custom_fig.renderText('ncOrtho')[:-3], flush=True)
    v = version('ncOrtho')
    print(f'Version: {v}\n', flush=True)
    ##########################################################################################
    # Parse arguments
    ##########################################################################################

    # Show help when no arguments are added.
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    else:
        args = parser.parse_args()
    
    # Check if computer provides the desired number of cores.
    available_cpu = mp.cpu_count()
    if args.cpu > available_cpu:
        raise ValueError('The provided number of CPU cores is higher than the number available on this system')
    else:
        cpu = args.cpu

    # mandatory
    mirnas = args.ncrna
    models = args.models
    output = os.path.realpath(args.output)
    query = args.query
    reference = args.reference

    # optional
    try:
        max_hits = int(args.maxcmhits)
    except TypeError:
        max_hits = None
    cm_cutoff = args.cm_cutoff
    checkCoorthref = args.checkCoorthologsRef
    cleanup = args.cleanup
    heuristic = (args.heuristic, args.heur_blast_evalue, args.heur_blast_length, args.sensitive_heuristic)
    msl = args.minlength
    refblast = args.refblast
    qblast = args.queryblast
    dust = args.dust.strip()

    ##########################################################################################
    # Starting argument checks
    ##########################################################################################

    # Test if optional query name was given
    if args.queryname:
        qname = args.queryname
    else:
        qname = '.'.join(query.split('/')[-1].split('.')[:-1])
    # make folder for query in output dir
    if not os.path.isdir(output):
        os.mkdir(output)
    if not output.split('/')[-1] == qname:
        output = f'{output}/{qname}'

    # Test if input files exist
    all_files = [mirnas, reference, query]
    for pth in all_files:
        if not os.path.isfile(pth):
            raise ValueError(f'{pth} is not a file')
    if not os.path.isdir(models):
        raise ValueError(f'Directory with covariance models does not exist at: {models}')

    # create symbolic links to guarantee writing permissions for pyfaidx index
    q_data = '{}/data'.format(output)
    if not os.path.isdir(q_data):
        os.makedirs(q_data)
    qlink = f'{q_data}/{qname}.fa'
    if not os.path.islink(qlink):
        os.symlink(query, qlink)

    # check if reference BLASTdb was given as input
    if refblast:
        # check if BLASTdb exists
        if not check_blastdb(refblast):
            raise ValueError('# Reference BLASTdb not found at: {}'.format(refblast))
    else:
        # check if refblast exists at location of reference genome
        if not check_blastdb(reference):
            refname = reference.split('/')[-1]
            refblast = f'{q_data}/{refname}'
            # check if BLASTdb exists already. Otherwise create it
            if not check_blastdb(refblast):
                make_blastndb(reference, refblast)
        else:
            refblast = reference
    # check if query BLASTdb was given as input (only used in heuristic mode)
    if qblast and heuristic[0]:
        # check if qblast exists
        if not check_blastdb(qblast):
            raise ValueError('# Query BLASTdb not found at: {}'.format(qblast))
    elif heuristic[0]:
        qblast = qlink
        # check if already exists
        if not check_blastdb(qlink):
            print('# Creating query Database', flush=True)
            make_blastndb(query, qlink)

    ###############################################################################################
    # Main
    ###############################################################################################

    # Print out query
    print('# Starting ncOrtho run for {}\n'.format(qname), flush=True)

    # Create miRNA objects from the list of input miRNAs.
    mirna_dict = rna_maker(mirnas, models, args.phmm, cpu)

    # setup logging
    shortlog = ['# miRNA\tSeconds\tStatus\n']
    logout = f'{output}/{qname}_extended.log'
    if os.path.isfile(logout):
        os.remove(logout)
    logger = logging.getLogger('ncortho')
    logger.setLevel(logging.DEBUG)
    fh = logging.FileHandler(logout)
    fh.setLevel(logging.DEBUG)
    logger.addHandler(fh)
    sucess_count = 0

    print('\n# Ortholog search', flush=True)
    mirna_orthologs = []
    for mirna in tqdm(mirna_dict, file=sys.stdout):
        st = time()
        restricted = False
        sys.stdout.flush()
        mirna_data = mirna_dict[mirna]
        # mirna objects for which no CM was found are empty
        if not mirna_data:
            et = time()
            elapsed_time = str(et - st)
            shortlog.append(f'{mirna}\t{elapsed_time}\tNo CM\n')
            continue
        logger.info(f'\n### {mirna}')

        # Create output folder, if not existent.
        if not heuristic[0] or not cleanup:
            outdir = '{}/{}'.format(output, mirna)
        else:
            outdir = '{}'.format(output)
        if not os.path.isdir(outdir):
            os.makedirs(outdir)

        # start cmsearch
        cm_results, exitstatus = model_search(
            mirna_data, cm_cutoff, cpu, msl, models, qlink, qblast, outdir, cleanup, heuristic, args.phmm
        )

        # Extract sequences for candidate hits (if any were found).
        if not cm_results:
            logger.info('# {} for {}'.format(exitstatus, mirna))
            et = time()
            elapsed_time = str(et - st)
            shortlog.append(f'{mirna}\t{elapsed_time}\t{exitstatus}\n')
            continue
        elif max_hits:
            if len(cm_results) > max_hits:
                logger.warning('# Maximum CMsearch hits reached. Restricting to best {} hits'.format(max_hits))
                cm_results = {k: cm_results[k] for k in list(cm_results.keys())[:max_hits]}
                restricted = True
        # get sequences for each CM result
        gp = GenomeParser(qlink, cm_results.values())
        candidates = gp.extract_sequences()
        # print(candidates)
        logger.info(f'Covariance model search successful, found {len(candidates)} ortholog candidate(s).')

        # Perform reverse BLAST test to verify candidates, stored in
        reblast_hits = {}
        for candidate in candidates:
            sequence = candidates[candidate]
            if list(sequence).count('N') >= 0.9 * len(sequence):
                continue
            # print('# Starting reverse blast for {}'.format(candidate))
            reblasthit = perform_reblast(
                sequence, refblast, cpu, outdir, candidate, mirna_data, dust, msl, cleanup, checkCoorthref
            )
            if reblasthit:
                reblast_hits[candidate] = reblasthit

        # Write output file if no candidate got accepted.
        if not reblast_hits:
            et = time()
            elapsed_time = str(et - st)
            logger.info(f'# None of the candidates for {mirna} could be verified.')
            if restricted:
                shortlog.append(f'{mirna}\t{elapsed_time}\tNo re-BLAST after restricting to {max_hits} CMsearch hits\n')
            else:
                shortlog.append(f'{mirna}\t{elapsed_time}\tNo re-BLAST\n')
            continue

        # Write output file if at least one candidate got accepted.
        nr_accepted = len(reblast_hits)
        if nr_accepted == 1:
            logger.info('# ncOrtho found 1 verified ortholog.')
            out_dict = reblast_hits
        else:
            logger.info(
                '# ncOrtho found {} potential co-orthologs.'
                .format(nr_accepted)
            )
            logger.info('# Evaluating distance between candidates to verify co-orthologs')
            rbp = ReBlastParser(mirna_data, reblast_hits)
            out_dict = rbp.verify_coorthologs(outdir)
            if len(out_dict) == 1:
                logger.info('ncOrtho found 1 verified ortholog')
            else:
                logger.info(f'# ncOrtho found {len(out_dict)} verified co-orthologs.')

        for hit in out_dict:
            cmres = list(cm_results[hit])
            cmres.insert(0, qname)
            cmres = [str(entry) for entry in cmres]
            header = '|'.join(cmres)
            mirna_orthologs.append('>{0}\n{1}\n'.format(header, out_dict[hit]))

        et = time()
        elapsed_time = str(round(et - st, 3))
        shortlog.append(f'{mirna}\t{elapsed_time}\tSUCESS\n')
        sucess_count += 1

    write_output(mirna_orthologs, f'{output}/{qname}_orthologs.fa')
    write_output(shortlog, f'{output}/{qname}_summary.log')
    # logging.basicConfig(filename=f'{output}/{qname}_extended.log', level=logging.DEBUG)

    print(f'\n### ncOrtho found orthologs for {sucess_count} out of {len(mirna_dict)} ncRNAs', flush=True)


if __name__ == "__main__":
    main()

